from IPython.core.display import display, HTML, JSON, Image, SVG, Markdown
from .types import t_notebook
from .manager import Manager
import os

class Api():
    naas = None
    role = t_notebook

    def __init__(self):
        self.manager = Manager()

    def current_raw(self):
        json_data = self.manager.get_naas()
        for item in json_data:
            if item['type'] == self.role:
                print(item)
    
    def currents(self):
        json_data = self.manager.get_naas()
        for item in json_data:
            kind = None
            if item['type'] == self.role:
                kind = f"callable with this url {self.manager.proxy_url(item['value'], 'start')} from anywhere"
                print(f"File ==> {item['path']} is {kind}")

    def add(self, path=None, params={}, silent=False):
        current_file = self.manager.get_path(path)
        if current_file is None:
            print('Missing file path in prod mode')
            return
        prod_path = self.manager.get_prod_path(current_file)
        token = self.manager.get_value(prod_path, self.role)
        if token is None:
            token = os.urandom(30).hex()
        url = self.manager.proxy_url(token, 'start')
        if not self.manager.notebook_path():
            print(
                f'No add done you are in already in naas folder\n')
            return url
        if silent is False:
            print(
                f'[Naas from Jupyter] => i have copied this {current_file} here: {prod_path} \n')
            print(
                f'You can now use it in Naas workspace with this url : {url} \n')
            display(HTML(f'<a href="{url}"">{url}</a>'))
            print(
                f'If you want to remove the api capability, just call .delete({path if path is not None else "" })) in this file')
        self.manager.add_prod({"type": self.role, "path": current_file, "params": params, "value": token}, silent)
        return url

    def respond_html(self, data):
        display(HTML(data, metadata={'naas_api': True}))

    def respond_json(self, data):        
        display(JSON(data, metadata={'naas_api': True}))

    def respond_image(self, data, filename, mimetype):
        display(Image(data, filename=filename, format=mimetype, metadata={'naas_api': True}))
        
    def respond_svg(self, data):
        display(SVG(data, metadata={'naas_api': True}))
        
    def respond_markdown(self, data):
        display(Markdown(data, metadata={'naas_api': True}))

    def respond_csv(self, data):
        display(HTML(empDfObj.to_html(), metadata={'naas_api': True, 'naas_type': 'csv'}))

    def get(self, path=None):
        current_file = self.manager.get_path(path)
        self.manager.get_prod(current_file)

    def get_output(self, path=None):
        current_file = self.manager.get_path(path)
        self.manager.get_output(current_file)

    def clear_output(self, path=None):
        current_file = self.manager.get_path(path)
        self.manager.clear_output(current_file)
                
    def get_history(self, path=None, histo=None):
        if not histo:
            print(
                f'No histo provided\n')
            return     
        current_file = self.manager.get_path(path)
        self.manager.get_history(current_file, histo)

    def list_history(self, path=None):
        current_file = self.manager.get_path(path)
        self.manager.list_history(current_file)
        
    def clear_history(self,  path=None, histo=None):
        current_file = self.manager.get_path(path)
        self.manager.clear_history(current_file, histo)

    def delete(self, path=None, all=False, silent=False):
        if not self.manager.notebook_path():
            print(
                f'No delete done you are in already in naas folder\n')
            return
        current_file = self.manager.get_path(path)
        self.manager.del_prod({"type": self.role, "path": current_file}, silent)
        if all is True:
            self.manager.clear_history(current_file)
            self.manager.clear_output(current_file) 
                
    def help(self):
        print(f'=== {type(self).__name__} === \n')
        print(f'.add(path, params) => add path to the prod {type(self).__name__} server\n')
        print(f'.delete(path) => delete path to the prod {type(self).__name__} server\n')
        print('.clear_history(path, histonumber) => clear history, history number and path are optionel, if you don\'t provide them it will erase full history of current file \n')
        print('.list_history(path) => list history, of a path or if not provided the current file \n')
        print('.get_history(histonumber, path) => get history file, of a path or if not provided the current file \n')
        print('.get(path) => get current prod file of a path, or if not provided the current file \n')
        print(f'.currents() => get current list of {type(self).__name__} prod file\n')
        print(f'.current_raw() => get json current list of {type(self).__name__} prod file\n')



from IPython.core.display import display, HTML
from .types import t_static
from .manager import Manager
import os

class Static():
    naas = None
    role = t_static

    def __init__(self):
        self.manager = Manager()

    def current_raw(self):
        json_data = self.manager.get_naas()
        for item in json_data:
            if item['type'] == self.role:
                print(item)
    
    def currents(self):
        json_data = self.manager.get_naas()
        for item in json_data:
            kind = None
            if item['type'] == self.role:
                kind = f"publicly gettable with this url {self.manager.proxy_url(item['value'])}"
                print(f"File ==> {item['path']} is {kind}")

    def get(self, path=None):
        if not self.manager.notebook_path():
            print(
                f'No get done you are in already in naas folder\n')
            return  
        current_file = self.manager.get_path(path)
        self.manager.get_prod(current_file)

    def get_history(self, histo, path=None):
        if not self.manager.notebook_path():
            print(
                f'No get history done you are in already in naas folder\n')
            return  
        current_file = self.manager.get_path(path)
        self.manager.get_history(current_file, histo)

    def list_history(self, path=None):
        current_file = self.manager.get_path(path)
        self.manager.list_history(current_file)
        
    def clear_history(self, path=None, histo=None):
        if not self.manager.notebook_path():
            print(
                f'No clear history done you are in already in naas folder\n')
            return  
        current_file = self.manager.get_path(path)
        self.manager.clear_history(current_file, histo)
                
    def add(self, path=None, params={}, silent=False, Force=False):
        current_file = self.manager.get_path(path)
        if current_file is None:
            print('Missing file path in prod mode')
            return
        prod_path = self.manager.get_prod_path(current_file)
        token = self.manager.get_value(prod_path, self.role)
        if token is None or Force is True:
            token = os.urandom(30).hex()
        url = self.manager.proxy_url(token)
        if not self.manager.notebook_path() and Force is False:
            print(
                f'No add done you are in already in naas folder\n')
            return url
        if silent is False:
            print(
                f'[Naas from Jupyter] => i have copied this {current_file} here: {prod_path} \n')
            print(
                f'You can now call from anywere this url : \n')
            display(HTML(f'<a href="{url}"">{url}</a>'))
            print(
                f'If you want to remove the api capability, just call .delete({path if path is not None else "" }) in this file')
        self.manager.add_prod({"type": self.role, "path": current_file, "params": params, "value": token}, silent)
        return url

    def delete(self, path=None, all=False, silent=False):
        if not self.manager.notebook_path():
            print(
                f'No delete done you are in already in naas folder\n')
            return        
        current_file = self.manager.get_path(path)
        self.manager.del_prod({"type": self.role, "path": current_file}, silent)
        if all is True:
            self.manager.clear_history(current_file)
            self.manager.clear_output(current_file) 
            
    def help(self):
        print(f'=== {type(self).__name__} === \n')
        print(f'.add(path, params) => add path to the prod {type(self).__name__} server\n')
        print(f'.delete(path) => delete path to the prod {type(self).__name__} server\n')
        print('.clear_history(path, histonumber) => clear history, history number and path are optionel, if you don\'t provide them it will erase full history of current file \n')
        print('.list_history(path) => list history, of a path or if not provided the current file \n')
        print('.get_history(path, histonumber) => get history file, of a path or if not provided the current file \n')
        print('.get(path) => get current prod file of a path, or if not provided the current file \n')
        print(f'.currents() => get current list of {type(self).__name__} prod file\n')
        print(f'.current_raw() => get json current list of {type(self).__name__} prod file\n')
    


from .proxy import encode_proxy_url
import pretty_cron
import requests
import base64
import json
import os

class Notifications():
    logger = None
    base_notif_url = os.environ.get('NOTIFICATIONS_API', '')
    
    def __init__(self, logger):
        self.logger = logger
    
    def send(self, uid, status, email, file_path, current_type):
        content = f"Your {file_path} accesible as {current_type} is {status}, check the Logs on your manager below :"
        status_url = f"{encode_proxy_url('assets')}/{status}.png"
        message_bytes = file_path.encode('ascii')
        base64_bytes = base64.b64encode(message_bytes)
        file_path_base64 = base64_bytes.decode('ascii')
        link_url = f"{encode_proxy_url('manager')}/?filter={file_path_base64}"
        try:
            data = {'subject': f'{current_type.capitalize()} {status}', 'email': email, 'content': content, 'image': status_url, 'link': link_url}
            req = requests.post(url=f"{self.base_notif_url}/send", json=data)
            req.raise_for_status()
            jsn = req.json()
            return jsn
        except Exception as err:
            self.logger.error(json.dumps(
                {'id': uid, 'type': 'notification error', 'error': str(err)}))

    def get_status_server(self):
        req = requests.get(url=f"{self.base_notif_url}/status")
        req.raise_for_status()
        jsn = req.json()
        return jsn
        
    def send_scheduler(self, uid, status, email, file_path, cron_str):
        cron_string = pretty_cron.prettify_cron(cron_str)
        content = f"Your {file_path} who run {cron_string} is {status}, check the Logs on your manager below :"
        status_url = f"{encode_proxy_url('assets')}/{status}.png"
        message_bytes = file_path.encode('ascii')
        base64_bytes = base64.b64encode(message_bytes)
        file_path_base64 = base64_bytes.decode('ascii')
        link_url = f"{encode_proxy_url('manager')}/?filter={file_path_base64}"
        try:
            data = {'subject': f'Sheduler {status}', 'email': email, 'content': content, 'image': status_url, 'link': link_url}
            req = requests.post(url=f"{self.base_notif_url}/send", json=data)
            req.raise_for_status()
            jsn = req.json()
            return jsn
        except Exception as err:
            self.logger.error(json.dumps(
                {'id': uid, 'type': 'notification error', 'error': str(err)}))


    def send_error(self, uid, error, file_path):
        email = os.environ.get('JUPYTERHUB_USER', None)
        if email is not None:
            content = f"Your {file_path} got Errors : {error},  during run N: {uid} \n Check the Logs on your manager below :"
            status_url = f"{encode_proxy_url('assets')}/down.png"
            message_bytes = file_path.encode('ascii')
            base64_bytes = base64.b64encode(message_bytes)
            file_path_base64 = base64_bytes.decode('ascii')
            link_url = f"{encode_proxy_url('manager')}/?filter={file_path_base64}"
            try:
                data = {'subject': 'Manager Error', 'email': email, 'content': content, 'image': status_url, 'link': link_url}
                req = requests.post(url=f"{self.base_notif_url}/send", json=data)
                req.raise_for_status()
                jsn = req.json()
                return jsn
            except Exception as err:
                self.logger.error(json.dumps(
                    {'id': uid, 'type': 'notification error', 'error': str(err)}))



# from gevent.lock import BoundedSemaphore
from asyncio import Semaphore
from naas.types import t_delete, t_add, t_skip, t_update, t_error, t_start
import pandas as pd
import datetime
import errno
import json
import os
import uuid

class Jobs():
    __storage_sem = None
    __path_user_files = None
    __logger = None
    __naas_folder = '.naas'
    __json_name = 'jobs.json'

    def __init__(self, logger, clean = False, init_data = []):
        self.__path_user_files = os.environ.get('JUPYTER_SERVER_ROOT', '/home/ftp')
        self.__path_naas_files = os.path.join(self.__path_user_files, self.__naas_folder)
        self.__json_secrets_path = os.path.join(self.__path_naas_files, self.__json_name)
        # self.__storage_sem = BoundedSemaphore(1)
        self.__storage_sem = Semaphore(1)
        self.__logger = logger
        if not os.path.exists(self.__path_naas_files):
            try:
                print('Init Naas folder Jobs')
                os.makedirs(self.__path_naas_files)
            except OSError as exc: # Guard against race condition
                print('__path_naas_files', self.__path_naas_files)
                if exc.errno != errno.EEXIST:
                    raise 
            except Exception as e:
                print('Exception', e)
        if not os.path.exists(self.__json_secrets_path) or clean:
            uid = str(uuid.uuid4())
            try:
                print('Init Job Storage', self.__json_secrets_path)
                self.__save(uid, init_data)
            except Exception as e:
                print('Exception', e)
                self.__logger.error(
                    {'id': uid, 'type': 'init_job_storage', "status": 'error', 'error': str(e)})          

    def __save(self, uid, data):
        try:
            with open(self.__json_secrets_path, 'w+') as f:
                f.write(json.dumps(data, sort_keys=True, indent=4).replace('NaN' , 'null'))
                f.close()
        except Exception as err:
            print('__save Exception', err)
            self.__logger.error({'id': str(uid), 'type': 'set_job_storage',
                                    "status": 'exception', 'filepath': self.__json_secrets_path, 'error': str(err)})

    def find_by_value(self, uid, value, target_type):
        data = self.list(uid)
        if (len(data) > 0):
            job_list = pd.DataFrame(data)
            cur_jobs = job_list[(job_list.type == target_type)
                            & (job_list.value == value)]
            cur_job = cur_jobs.to_dict('records')
            if (len(cur_job) == 1):
                return cur_job[0]
        return None

    def find_by_path(self, uid, filepath, target_type):
        data = self.list(uid)
        if (len(data) > 0):
            job_list = pd.DataFrame(data)
            cur_jobs = job_list[(job_list.type == target_type)
                            & (job_list.path == filepath)]
            cur_job = cur_jobs.to_dict('records')
            if (len(cur_job) == 1):
                return cur_job[0]
        return None

    def is_running(self, uid, notebook_filepath, target_type):
        cur_job = self.find_by_path(uid, notebook_filepath, target_type)
        if cur_job:
            status = cur_job.get('status', None)
            if (status and status == t_start):
                return True
        return False

    def list(self, uid):
        data = []
        try:
            with open(self.__json_secrets_path, 'r') as f:
                data = json.load(f)
                f.close()
        except Exception as err:
            print('cannot open')
            self.__logger.error({'id': uid, 'type': 'get_job_storage',
                                    "status": 'exception', 'filepath': self.__json_secrets_path, 'error': str(err)})
            data = []
        return data
    
    async def update(self, uid, path, target_type, value, params, status, runTime = 0):
        await self.__storage_sem.acquire()
        data = None
        res = t_error
        try:
            if (len(self.list(uid)) != 0):
                df = pd.DataFrame(self.list(uid))
            else:
                df = pd.DataFrame(columns=['id', 'type', 'value', 'path', 'status', 'params', 'lastUpdate', 'lastRun', 'totalRun']) 
            res = status
            cur_elem = df[(df.type == target_type) & (df.path == path)]
            now = datetime.datetime.now()
            dt_string = now.strftime("%Y-%m-%d %H:%M:%S")
            if (len(cur_elem) == 1):
                if (status == t_delete):
                    self.__logger.info({'id': uid, 'type': target_type, 'value': value, 'status': t_delete,
                                'path': path, 'params': params})
                    df = df.drop(cur_elem.index)
                else:
                    self.__logger.info({'id': uid, 'type': target_type, 'value': value, 'status': t_update,
                            'path': path, 'params': params})
                    index = cur_elem.index[0]
                    df.at[index,'id'] = uid
                    df.at[index,'status'] = status
                    df.at[index,'value'] = value
                    df.at[index,'params'] = params
                    df.at[index,'lastUpdate'] = dt_string
                    if runTime > 0 and status != t_add:
                        df.at[index,'lastRun'] = runTime
                        df.at[index,'totalRun'] = runTime + (df.at[index,'totalRun'] if df.at[index,'totalRun'] else 0)
                    elif status == t_add:
                        df.at[index,'lastRun'] = 0
                        df.at[index,'totalRun'] = 0                    
                    res = t_update
            elif (status == t_add and len(cur_elem) == 0):
                self.__logger.info({'id': uid, 'type': target_type, 'value': value, 'status': t_update,
                        'path': path, 'params': params})
                new_row = [{'id': uid,'type': target_type, 'value': value, 'status': t_add,
                            'path': path, 'params': params, 'lastRun': runTime, 'totalRun': runTime, 'lastUpdate':  dt_string}]
                df_new = pd.DataFrame(new_row)
                df = pd.concat([df, df_new], axis=0)
            else:
                res = t_skip
            data = df.to_dict('records')
            if (res != t_skip):
                self.__save(uid, data)
        except Exception as e:
            print('cannot update', e)
            self.__logger.error({'id': uid, 'type': target_type, 'value': value, 'status': t_error,
                        'path': path, 'params': params, 'error': str(e)})
        self.__storage_sem.release()
        return {"status": res, "data": data}





from .types import t_delete, t_add, t_skip, t_update, t_error, t_static, t_notebook, t_scheduler, t_dependency
from notebook import notebookapp
from .runner.proxy import encode_proxy_url
from shutil import copy2
import ipykernel
import requests
import datetime
import urllib
import errno
import json
import os
import re

class Manager():
    __port = os.environ.get('NAAS_RUNNER_PORT', 5000)
    __local_api = f'http://localhost:{__port}'
    __base_ftp_path = None
    __public_url = None
    __jup_user = None
    __jup_token = None
    __production_path = None
    __folder_name = '.naas'
    __json_files = 'jobs.json'
    __readme_name = 'README.md'
    __json_files_path = None
    __readme_path = None

    def __init__(self):
        self.__base_ftp_path = os.environ.get('JUPYTER_SERVER_ROOT', '/home/ftp')
        self.__public_url = os.environ.get('PUBLIC_DATASCIENCE', '')
        self.__jup_user = os.environ.get('JUPYTERHUB_USER', '')
        self.__jup_token = os.environ.get('JUPYTERHUB_API_TOKEN', '')
        self.__production_path = os.path.join(self.__base_ftp_path, self.__folder_name)
        self.__json_files_path = os.path.join(self.__production_path, self.__json_files)
        self.__readme_path = os.path.join(self.__production_path, self.__readme_name)
        if not os.path.exists(self.__production_path):
            try:
                os.makedirs(self.__production_path)
            except OSError as exc: # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        try:
            with open(self.__readme_path, 'w+') as readme:
                readme.write('Welcome NAAS')
                readme.close()
        except OSError as exc: # Guard against race condition
            if exc.errno != errno.EEXIST:
                raise 

    def notebook_path(self):
        """Returns the absolute path of the Notebook or None if it cannot be determined
        NOTE: works only when the security is token-based or there is also no password
        """
        try:
            connection_file = os.path.basename(ipykernel.get_connection_file())
            kernel_id = connection_file.split('-', 1)[1].split('.')[0]

            for srv in notebookapp.list_running_servers():
                try:
                    base_url = f'{self.__public_url}/user/{self.__jup_user}/api/sessions'
                    req = urllib.request.urlopen(f'{base_url}?token={self.__jup_token}')
                    sessions = json.load(req)
                    for sess in sessions:
                        if sess['kernel']['id'] == kernel_id:
                            return os.path.join(srv['notebook_dir'],sess['notebook']['path'])
                except:
                    pass  # There may be stale entries in the runtime directory 
        except:
            pass
        return None

    def get_path(self, path):
        if path is not None:
            return os.path.abspath(os.path.join(os.getcwd(), path))
        else:
            return self.notebook_path()

    def proxy_url(self, token, endpoint=None):
        public_url = encode_proxy_url(token)
        if endpoint:
            public_url = f"{public_url}/{endpoint}"
        return public_url

    def get_prod_path(self, path):
        new_path = path.replace(self.__base_ftp_path, self.__production_path)
        return new_path

    def __copy_file_in_dev(self, path):
        new_path = self.get_prod_path(path)
        if not os.path.exists(new_path):
            raise Exception(f"file doesn't exist {new_path}")
        if not os.path.exists(os.path.dirname(path)):
            try:
                os.makedirs(os.path.dirname(path))
            except OSError as exc: # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        dev_dir = os.path.dirname(path)
        dev_finename = os.path.basename(path)
        secure_path = os.path.join(dev_dir, f'prod_{dev_finename}')
        try:
            copy2(new_path, secure_path)
        except Exception as e:
            raise Exception(f"Cannot copied here {secure_path}, file probabily exist {path} {str(e)}")
        print(f'File copied here {secure_path}')
        return secure_path
    
    def __copy_file_in_prod(self, path):
        new_path = self.get_prod_path(path)
        prod_dir = os.path.dirname(new_path)
        prod_finename = os.path.basename(new_path)
        history_filename = f'{datetime.datetime.now().strftime("%Y%m%d%H%M%S%f")}_{prod_finename}'
        history_path = os.path.join(prod_dir, history_filename)
        if not os.path.exists(path):
            raise Exception(f"file doesn't exist {path}")
        if not os.path.exists(os.path.dirname(new_path)):
            try:
                os.makedirs(os.path.dirname(new_path))
            except OSError as exc: # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        if os.path.exists(new_path):
            os.remove(new_path)
        copy2(path, new_path)
        copy2(path, history_path)
        return new_path

    def __del_file_in_prod(self, path):
        if path.find(self.__production_path) == -1:
            raise Exception(f"Cannot delte file {path} it's in other folder than {self.__production_path}")
        if os.path.exists(path):
            os.remove(path)

    def get_naas(self):
        naas_data = []
        try:
            with open(self.__json_files_path, 'r') as f:
                naas_data = json.load(f)
                f.close()
        except:
            naas_data = [] 
        return naas_data
    
    def get_value(self, path, obj_type):
        json_data = self.get_naas()
        value = None
        for item in json_data:
            if (item['type'] == obj_type and item['path'] == path):
                value = item['value']
        return value

    def is_already_use(self, obj):
        json_data = self.get_naas()
        already_use = False
        for item in json_data:
            if ("type" in obj and "value" in obj
                and ( obj['type'] == t_static or obj['type'] == t_notebook )
                and item['type'] == obj['type']
                and item['value'] == obj['value']):
                already_use = True
        return already_use

    def get_prod(self, path):
        self.__copy_file_in_dev(path)
                
    def get_output(self, path):
        filename = os.path.basename(path)
        dirname = os.path.dirname(path)
        out_path = os.path.join(dirname, f'out_{filename}')
        self.__copy_file_in_dev(out_path)
        
    def clear_output(self, path):
        filename = os.path.basename(path)
        dirname = os.path.dirname(path)
        out_path = os.path.join(dirname, f'out_{filename}')
        print(f'Delete {out_path}')
        os.remove(out_path)   
        
    def list_history(self, path):
        prod_path = self.get_prod_path(path)
        filename = os.path.basename(path)
        dirname = os.path.dirname(prod_path)
        print(f'Avaliable :\n')
        for ffile in os.listdir(dirname):
            if ffile.endswith(filename) and ffile != filename and not ffile.startswith('out'):
                histo = ffile.replace(filename, '')
                histo = histo.replace('_', '')
                print(histo + '\n')
                
    def get_history(self, path, histo):
        filename = os.path.basename(path)
        dirname = os.path.dirname(path)
        path_histo = os.path.join(dirname, f'{histo}_{filename}')
        self.__copy_file_in_dev(path_histo)
                
    def clear_history(self, path, histo=None):
        prod_path = self.get_prod_path(path)
        filename = os.path.basename(path) if not histo else f'{histo}_{os.path.basename(path)}'
        dirname = os.path.dirname(prod_path)
        for ffile in os.listdir(dirname):
            if not ffile.startswith('out_') and ((histo and filename == ffile) or (not histo and ffile.endswith(filename) and filename != ffile)):
                tmp_path = os.path.join(dirname, ffile)
                print(f'Delete {tmp_path}')
                os.remove(tmp_path)    

    def add_prod(self, obj, silent):
        if "type" in obj and "path" in obj and "params" in obj and "value" in obj:
            dev_path = obj.get('path')
            obj['path'] = self.get_prod_path(obj.get('path'))
            self.__copy_file_in_prod(dev_path)
            r = requests.post(f"{self.__local_api}/v1/jobs", json={**obj, **{'status': t_add}})
            if not silent:
                print(f'{r.json()["status"]} ==> {obj}')
            return obj
        else:
            raise Exception('obj should have all keys ("type","path","params","value" )')

    def del_prod(self, obj, silent):
        if "type" in obj and "path" in obj:
            obj['path'] = self.get_prod_path(obj.get('path'))
            self.__del_file_in_prod(obj['path'])
            r = requests.post(f"{self.__local_api}/v1/jobs", json={**obj, **{'status': t_delete}})
            if not silent:
                print(f'{r.json()["status"]} ==> {obj}')
            return obj
        else:
            raise Exception('obj should have keys ("type","path")')



import pandas as pd
import traceback
from naas.types import t_add
import logging
import errno
import uuid
import json
import csv
import io
import os
import datetime as dt


class CsvFormatter(logging.Formatter):
    """
    Custom logger for the csv logs
    """
    converter=dt.datetime.fromtimestamp
    
    def __init__(self, set_headers=False):
        super().__init__(datefmt='%Y-%m-%d %H:%M:%S.%f')
        self.output = io.StringIO()
        self.writer = csv.writer(self.output, delimiter=';', quoting=csv.QUOTE_ALL)
        # if set_headers:
        #     print('write headers')
        #     self.writer.writerow(['asctime', 'levelname', 'name', 'message'])

    def formatTime(self, record, datefmt=None):
        ct = self.converter(record.created)
        if datefmt:
            s = ct.strftime(datefmt)
        else:
            t = ct.strftime("%Y-%m-%d %H:%M:%S")
            s = "%s,%03d" % (t, record.msecs)
        return s
    
    def format(self, record):
        record.asctime = self.formatTime(record, self.datefmt)
        self.writer.writerow(
            [record.asctime, record.levelname, record.name, record.msg])
        data = self.output.getvalue()
        self.output.truncate(0)
        self.output.seek(0)
        return data.strip()

class Logger():
    __log = None
    __name = 'naas_logger'
    __logs_filename = 'logs.csv'
    __naas_folder = '.naas'
    __path_user_files = None
    
    def __init__(self, reset=False):
        self.__path_user_files = os.environ.get('JUPYTER_SERVER_ROOT', '/home/ftp')   
        self.__path_naas_files = os.path.join(self.__path_user_files, self.__naas_folder)
        self.__path_logs_file = os.path.join(self.__path_naas_files, self.__logs_filename) 
        if not os.path.exists(self.__path_naas_files):
            try:
                print('Init Naas folder Logger')
                os.makedirs(self.__path_naas_files)
            except OSError as exc: # Guard against race condition
                if exc.errno != errno.EEXIST:
                    raise
        if reset or not os.path.exists(self.__path_logs_file):
            with open(self.__path_logs_file, 'w') as fp:
                separator = ';'
                fp.write(f"{separator.join(['asctime', 'levelname', 'name', 'message'])}\n")
                pass
        self.__log = logging.getLogger(self.__name)
        handler = logging.FileHandler(self.__path_logs_file, "a")
        handler.setFormatter(CsvFormatter())
        self.__log.addHandler(handler)
        logging.basicConfig(level=logging.INFO)
        
    def info(self, data):
        self.__log.info(json.dumps(data))

    def error(self, data):
        self.__log.error(json.dumps(data))

    def clear(self):
        os.remove(self.__path_logs_file)

    def get_file_path(self):
        return self.__path_logs_file
    
    def list(self, uid: str, skip: int = 0, limit: int = 0, search: str = '', filters: list = []):
        df = None
        try:
            df = pd.read_csv(self.__path_logs_file, sep=';', index_col=0)
            with open(self.__path_logs_file, 'r') as fp:
                print('fp', fp.read())
            print(df)
            df1 = pd.DataFrame(df.pop('message').apply(
                pd.io.json.loads).values.tolist(), index=df.index)
            df = pd.concat([df1, df], axis=1, sort=False)
            if len(filters) > 0:
                df = df[df.type.isin(filters)]
            totalRecords = len(df.index)
            if search and search != '':
                idx = df.apply(lambda ts: any(ts.str.contains(
                    search, na=False, regex=False)), axis=1)
                df = df[idx]
            if skip > 0:
                df = df.iloc[skip:]
            if limit > 0:
                df = df[:limit]
            df = df.reset_index()
            return {'data': json.loads(df.to_json(orient='records')), 'totalRecords': totalRecords}
        except Exception as e:
            tb = traceback.format_exc()
            print('list logs', e, tb)
            return {'data': [], 'totalRecords': 0}
            
    



from sanic.response import json, redirect, stream
from .proxy import escape_kubernet
from naas.types import t_notebook, t_scheduler, t_error, t_health
import papermill as pm
import traceback
import asyncio
import time
import os
import csv
import io
import bs4

class Notebooks():
    __logger = None
    __port = None
    __notif = None
    __loop = None
    __api_internal = None
    
    def __init__(self, logger, loop, notif):
        self.__port = int(os.environ.get('NAAS_RUNNER_PORT', 5000))
        self.__user = os.environ.get('JUPYTERHUB_USER', 'joyvan@naas.com')
        self.__single_user_api_path = os.environ.get('SINGLEUSER_PATH', '.jupyter-single-user.dev.svc.cluster.local')
        self.__api_internal = f"http://jupyter-{escape_kubernet(self.__user)}{self.__single_user_api_path}:{self.__port}/"
        self.__loop = loop
        self.__logger = logger
        self.__notif = notif
        
    def response(self, uid, res, duration, params):
        next_url = params.get('next_url', None)
        if next_url is not None:
            if "http" not in next_url:
                next_url = f'{self.__api_internal}{next_url}'
            self.__logger.info(
                {'id': uid, 'type': t_notebook, 'status': 'next_url', 'url': next_url})
            return redirect(next_url, code=302)
        else:
            res_data = self.__get_res(res)
            if res_data and res_data.get('type') == 'application/json':
                return json(res_data.get('data'))
            elif res_data and res_data.get('type') == 'text/csv':
                csv_data = self.__convert_csv(res_data.get('data'))
                print('csv_data', csv_data)
                async def streaming_fn(res):
                    await res.write(csv_data)
                return stream(streaming_fn, content_type=res_data.get('type'))
            elif res_data:
                async def streaming_fn(res):
                    await res.write(res_data.get('data'))
                return stream(streaming_fn, content_type=res_data.get('type'))
        return json({'id': uid, "status": "Done", "time": duration})

    def __convert_csv(self, data):
        soup = bs4.BeautifulSoup(data)
        output = []
        for table_num, table in enumerate(soup.find_all('table')):
            csv_string = io.StringIO()
            csv_writer = csv.writer(csv_string, delimiter=';', quoting=csv.QUOTE_ALL)
            for tr in table.find_all('tr'):
                row = [''.join(cell.stripped_strings) for cell in tr.find_all(['td', 'th'])]
                csv_writer.writerow(row)
            table_attrs = dict(num=table_num)
            output.append((csv_string.getvalue(), table_attrs))
        return output[0][0]

    def __get_res(self, res):
        cells = res.get('cells')
        result = None
        result_type = None
        for cell in cells:
            outputs = cell.get('outputs', [])
            for output in outputs:
                metadata = output.get('metadata', [])
                data = output.get('data', dict())
                for meta in metadata:
                    if metadata[meta].get('naas_api'):
                        if (data.get('application/json')):
                            result_type = 'application/json'
                        elif (data.get('text/html') and metadata[meta].get('naas_type') == 'csv'):
                            result_type = 'text/csv'
                            result = data.get('text/html')
                            break
                        elif (data.get('text/html')):
                            result_type = 'text/html'
                            result = data.get('text/html')
                        elif (data.get('image/jpeg')):
                            result_type = 'image/jpeg'
                            result = data.get('image/jpeg')
                        elif (data.get('image/png')):
                            result_type = 'image/png'
                            result = data.get('image/png')
                        elif (data.get('image/svg+xml')):
                            result_type = 'image/svg+xml'
                            result = data.get('image/svg+xml')
                        if result_type is not None:
                            result = data.get(result_type)
        if result is None or result_type is None:
            return None
        return {'type': result_type, 'data': result}


    async def exec(self, uid, job):
        value = job.get('value', None)
        current_type = job.get('type', None)
        file_filepath = job.get('path')
        if not os.path.exists(file_filepath):
            err = 'file not found'
            self.__logger.error({'id': uid, 'type': 'filepath', "status": t_error,
                                    'filepath': file_filepath, 'error': err})
            return {"error": err}
        file_dirpath = os.path.dirname(file_filepath)
        file_filename = os.path.basename(file_filepath)
        file_filepath_out = os.path.join(file_dirpath, f'out_{file_filename}')
        params = job.get('params', dict())
        notif_down = params.get('notif_down', None)
        notif_up = params.get('notif_up', None)
        params['run_uid'] = uid
        start_time = time.time()
        res = None
        try:
            res = pm.execute_notebook(
                input_path=file_filepath,
                output_path=file_filepath_out,
                progress_bar=False,
                cwd=file_dirpath,
                parameters=params
            )
        except pm.PapermillExecutionError as err:
            self.__logger.error({'id': uid, 'type': 'PapermillExecutionError', "status": t_error,
                                    'filepath': file_filepath, 'output_filepath': file_filepath_out, 'error': str(err)})
            res = {"error": str(err)}
        except:
            tb = traceback.format_exc()
            res = {'error': str(tb)}
            self.__logger.error({'id': uid, 'type': 'Exception', "status": t_error,
                                    'filepath': file_filepath, 'output_filepath': file_filepath_out, 'error': str(tb)})
        if (not res):
            res = {'error': 'Unknow error'}
            self.__logger.error({'id': uid, 'type': 'Exception', "status": t_error,
                                    'filepath': file_filepath, 'output_filepath': file_filepath_out, 'error': res['error']})
        else:
            self.__logger.error({'id': uid, 'type': 'Exception', "status": t_health,
                                    'filepath': file_filepath, 'output_filepath': file_filepath_out})
        res['duration'] = (time.time() - start_time)
        if(res.get('error')):
            self.__notif.send_error(uid, str(res.get('error')), file_filepath)
            if (notif_down and current_type == t_scheduler):
                self.__notif.send_scheduler(uid, 'down', notif_down, file_filepath, value)
            elif (notif_down):
                self.__notif.send(uid, 'down', notif_down, file_filepath, current_type)
        elif (notif_up and current_type == t_scheduler):
            self.__notif.send_scheduler(uid, 'up', notif_down, file_filepath, value)
        elif (notif_up):
            self.__notif.send(uid, 'up', notif_up, file_filepath, current_type)
        return res



from base64 import b64encode
import escapism
import string
import os

_docker_safe_chars = set(string.ascii_letters + string.digits)
_docker_escape_char_kubernet = "-"
_docker_escape_char_docker = "_"

def escape_kubernet(s):
    """Escape a string to kubernet-safe characters"""
    return escapism.escape(
        s,
        safe=_docker_safe_chars,
        escape_char=_docker_escape_char_kubernet,
    )
    
def escape_docker(s):
    """Escape a string to docker-safe characters"""
    return escapism.escape(
        s,
        safe=_docker_safe_chars,
        escape_char=_docker_escape_char_docker,
    )

def encode_proxy_url(token=''):
    client = os.environ.get('JUPYTERHUB_USER', '')
    base_public_url = os.environ.get('PUBLIC_PROXY_API', '')
    clientEncoded = escape_kubernet(client)
    message_bytes = clientEncoded.encode('ascii')
    base64_bytes = b64encode(message_bytes)
    username_base64 = base64_bytes.decode('ascii')
    return f"{base_public_url}/{username_base64}/{token}"



from sentry_sdk.integrations.sanic import SanicIntegration
from .controllers.scheduler import SchedulerController
from .controllers.static import StaticController
from .controllers.notebooks import NbController
from .controllers.jobs import JobsController
from .controllers.logs import LogsController
from sanic_openapi import swagger_blueprint
from .controllers.env import EnvController
from .notifications import Notifications
from .proxy import escape_kubernet
from .scheduler import Scheduler
from .notebooks import Notebooks
from daemonize import Daemonize
from naas.types import t_main
from datetime import datetime
from .logger import Logger
from sanic import Sanic
from .jobs import Jobs
import sentry_sdk
import getpass
import uuid
import os

class Runner():
    # Declare semaphore variable
    __name = 'naas_runner'
    __naas_folder = '.naas'
    __tasks_sem = None
    # Declare path variable
    __path_lib_files = os.path.dirname(os.path.abspath(__file__))
    __path_user_files = None
    __port = 5000
    __single_user_api_path = None
    __html_files = 'html'
    __assets_files = 'assets'
    __manager_index = 'manager.html'
    __log_filename = 'logs.csv'
    __tasks_files = 'jobs.json'
    __info_file = 'info.json'
    __app = None
    __http_server = None
    __notif = None
    __jobs = None
    __scheduler = None
    __daemon = None
    __sentry = None
    __logger = None
    __testing = False
    tz = None
    user = None
    public_url = None
    proxy_url = None
            
    def __init__(self):
        self.__path_user_files = os.environ.get('JUPYTER_SERVER_ROOT', '/home/ftp')
        self.__port = int(os.environ.get('NAAS_RUNNER_PORT', 5000))
        self.__user = os.environ.get('JUPYTERHUB_USER', 'joyvan@naas.com')
        self.__public_url = os.environ.get('PUBLIC_DATASCIENCE', f'http://localhost:{self.__port}')
        self.__proxy_url = os.environ.get('PUBLIC_PROXY_API', 'http://localhost:5002')
        self.__tz = os.environ.get('TZ', 'Europe/Paris')
        self.__path_naas_files = os.path.join(self.__path_user_files, self.__naas_folder)
        self.__path_html_files = os.path.join(self.__path_lib_files, self.__html_files)
        self.__path_manager_index = os.path.join(self.__path_html_files, self.__manager_index)
        self.__path_pidfile = os.path.join(self.__path_naas_files, f'{self.__name}.pid')

    def __main(self, debug=False):
        uid = str(uuid.uuid4())
        self.__logger.info(
            {'id': uid, 'type': t_main, "status": 'start API'})
        self.__app.run(host="0.0.0.0", port=self.__port, debug=debug, access_log=debug)

    def __version(self):
        version = None
        try:
            with open(os.path.join(self.__path_lib_files, self.__info_file), 'r') as json_file:
                version = json.load(json_file)
        except:
            version = {'error': 'cannot get info.json'}
        return version

    async def initialize_before_start(self, app, loop):
        self.__nb = Notebooks(self.__logger, loop, self.__notif)
        self.__app.add_route(NbController.as_view(self.__logger, self.__jobs, self.__nb), '/notebooks/<token>')
        self.__scheduler = Scheduler(self.__logger, self.__jobs, loop)
        self.__app.add_route(SchedulerController.as_view(self.__logger, self.__scheduler), '/scheduler/<mode>')
        self.__scheduler.start()

    def kill(self):
        if os.path.exists(self.__path_pidfile):
            try:
                with open(self.__path_pidfile, 'r') as f:
                    pid = f.read()
                    print('kill')
                    os.system(f'kill {pid}')
                    print(f'Deamon killed pid {pid}')
                    f.close()
            except Exception:
                print('No Deamon running')

    def init_app(self):
        self.__logger = Logger()
        self.__notif = Notifications(self.__logger)
        self.__jobs = Jobs(self.__logger)
        self.__app = Sanic(__name__)
        self.__app.register_listener(self.initialize_before_start, 'before_server_start')
        self.__app.add_route(EnvController.as_view(self.__logger, self.__version, self.__user, self.__public_url, self.__proxy_url, self.__tz), '/env')
        self.__app.add_route(LogsController.as_view(self.__logger), '/logs')
        self.__app.add_route(JobsController.as_view(self.__logger, self.__jobs), '/jobs')
        self.__app.add_route(StaticController.as_view(self.__logger,  self.__jobs, self.__path_lib_files), '/static/<token>')
        self.__app.static('/', self.__path_manager_index, name='manager.html')
        self.__app.blueprint(swagger_blueprint)
        return self.__app
                
    def start(self, deamon=True, port=None, debug=True):
        user = getpass.getuser()
        if (user != self.__user):
            raise Exception(f"{user} not autorized, use {self.__user} instead")
        self.kill()
        if port:
            self.__port = port
        self.init_app()
        if deamon:
            self.__sentry = sentry_sdk.init(
                dsn="https://28c6dea445f741a7a0c5e4db4df88df4@o448748.ingest.sentry.io/5430604",
                traces_sample_rate=1.0,
                environment=escape_kubernet(self.__user),
                integrations=[SanicIntegration()]
            )
            self.__daemon = Daemonize(app=self.__name, pid=self.__path_pidfile, action=self.__main)
            print('Start Runner Deamon Mode')
            self.__daemon.start()
        else:
            print('Start Runner front Mode')
            try:
                self.__main(debug)
            except KeyboardInterrupt:
                print('Shutdown server')
                sys.exit()
    


from .refresh import Refresh
from .api import Api
from .static import Static
from .dependency import Dependency
from IPython.core.display import display, HTML
from .secret import Secret
from .runner import Runner
from .runner.proxy import encode_proxy_url
import requests
import json
import os

''' Naas code '''

__location__ = os.getcwd()
refresh = Refresh
secret = Secret
runner = Runner
api = Api
static = Static
dependency = Dependency

def version():
    try:
        with open(os.path.join(__location__, 'info.json'), 'r') as json_file:
            version = json.load(json_file)
            return version
    except:
        return {'error': 'version error'}

def manager():
    public_url = f'{encode_proxy_url()}/manager'
    print('You can check your current tasks list here :')
    display(HTML(f'<a href="{public_url}"">Manager</a>'))

def refresh_status():
    req = requests.get(url='localhost:5000/scheduler')
    req.raise_for_status()
    jsn = req.json()
    print(jsn) 
    return jsn

def refresh_pause():
    req = requests.get(url='localhost:5000/scheduler/pause')
    req.raise_for_status()
    jsn = req.json()
    print(jsn) 
    return jsn
    
def refresh_resume():
    req = requests.get(url='localhost:5000/scheduler/resume')
    req.raise_for_status()
    jsn = req.json()
    print(jsn) 
    return jsn

__version__ = f"Nass {version()}"

def welcome():
    print('[Nass from JupyterHub] => Welcome to your ETL, read the documentation')
    print('Do naas.help() to get more info on what you can do.\n')

def help():
    print('naas.version() => Get the current version\n')
    print('naas.manager() => Get the url of the manager\n')
    print('naas.refresh => Get the notebook refresh driver\n')
    print('naas.api => Get the notebook api driver\n')
    print('naas.static => Get the file sharing driver\n')
    print('naas.dependency => Get the file dependency driver\n')



t_notebook = 'api'
t_static = 'static'
t_task = 'tasks'
t_dependency = 'dependency'
t_scheduler = 'scheduler'

t_add = 'installed'
t_delete = 'delete'
t_update = 'edited'
t_start = 'started'
t_skip = 'skiped'
t_error = 'error'
t_health = 'healthy'
t_main = 'main'


# from apscheduler.schedulers.background import BackgroundScheduler
from naas.types import t_scheduler, t_start, t_main, t_health, t_error
from apscheduler.schedulers.asyncio import AsyncIOScheduler
import apscheduler.schedulers.base
import asyncio
import traceback
import datetime
import time
import uuid
import sys

class Scheduler():
    __scheduler = None
    __logger = None
    __jobs = None

    def __init__(self, logger, jobs, loop):
        self.__logger = logger
        self.__jobs = jobs
        # self.__scheduler = BackgroundScheduler()
        self.__scheduler = AsyncIOScheduler({'event_loop': loop})
        self.state = self.__scheduler.state

    def status(self):
        if (self.__scheduler.state == apscheduler.schedulers.base.STATE_RUNNING):
            return 'running'
        elif (self.__scheduler.state == apscheduler.schedulers.base.STATE_PAUSED):
            return 'paused'

    def start(self):
        if (self.__scheduler.state != apscheduler.schedulers.base.STATE_RUNNING):
            self.__scheduler.add_job(func=self.__scheduler_function,
                            trigger="interval", seconds=60, max_instances=10)
            self.__scheduler.start()
            uid = str(uuid.uuid4())
            self.__logger.info({'id': uid, 'type': t_main, "status": 'start SCHEDULER'})

    def __scheduler_greenlet(self, main_uid, current_time, task):
        value = task.get('value', None)
        current_type = task.get('type', None)
        file_filepath = task.get('path')
        params = task.get('params', dict())
        uid = str(uuid.uuid4())
        running = self.__jobs.is_running(uid, file_filepath, current_type)
        if current_type == t_scheduler and value is not None and pycron.is_now(value, current_time) and not running:
            self.__logger.info(
                {'main_id': str(main_uid), 'id': uid, 'type': t_scheduler, 'status': t_start, 'filepath': file_filepath})
            self.__jobs.update(uid, file_filepath, t_scheduler, value, params, t_start)
            res = self.__exec_job(uid, task)
            if (res.get('error')):
                self.__logger.error({'main_id': str(main_uid), 'id': uid, 'type': t_scheduler, 'status': t_error,
                                        'filepath': file_filepath, 'duration': res.get('duration'), 'error': str(res.get('error'))})
                self.__jobs.update(uid, file_filepath, t_scheduler, value,
                            params, t_error, res.get('duration'))
                return
            self.__logger.info(
                {'main_id': str(main_uid), 'id': uid, 'type': t_scheduler, 'status': t_health, 'filepath': file_filepath, 'duration': res.get('duration')})
            self.__jobs.update(uid, file_filepath, t_scheduler, value, params,
                        t_health, res.get('duration'))
            
    async def __scheduler_function(self):
        # Create unique for scheduling step (one step every minute)
        main_uid = str(uuid.uuid4())
        try:
            current_time = datetime.datetime.now()
            all_start_time = time.time()
            greelets = []
            # Write self.__scheduler init info in self.__logger.write
            self.__logger.info({'id': main_uid, 'type': t_scheduler,
                                    'status': t_start})
            for job in  self.__jobs.list(main_uid):
                g = asyncio.ensure_future(self.__scheduler_greenlet(main_uid, current_time, job))
                # g = gevent.spawn(self.__scheduler_greenlet,
                #                 main_uid, current_time, job)
                greelets.append(g)
            await asyncio.gather(*greelets)
            # gevent.joinall(greelets)
            durationTotal = (time.time() - all_start_time)
            self.__logger.info({'id': main_uid, 'type': t_scheduler, 'status': t_health, 'duration': durationTotal})
        except Exception as e:
            tb = traceback.format_exc()
            durationTotal = (time.time() - all_start_time)
            self.__logger.error({'id': main_uid, 'type': t_scheduler, 'status': t_error,
                                    'duration': durationTotal, 'error': str(e), 'trace': tb})
        except:
            tb = traceback.format_exc()
            durationTotal = (time.time() - all_start_time)
            self.__logger.error({'id': main_uid, 'type': t_scheduler, 'status': t_error,
                                    'duration': durationTotal, 'error': str(sys.exc_info()[0]), 'trace': tb})

